/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
#include "compat_memstream.h"

static size_t dev_hash(const void *x, size_t tablesize)
{
	const struct mog_dev *dev = x;

	return dev->st_dev % tablesize;
}

static bool dev_cmp(const void *a, const void *b)
{
	const struct mog_dev *dev_a = a;
	const struct mog_dev *dev_b = b;

	return dev_a->st_dev == dev_b->st_dev;
}

static void svc_init_dev_hash(struct mog_svc *svc)
{
	if (svc->by_st_dev) {
		hash_clear(svc->by_st_dev);
		return;
	}
	svc->by_st_dev = hash_initialize(7, NULL, dev_hash, dev_cmp, free);
	if (!svc->by_st_dev) mog_oom();
}

static int svc_scandev(struct mog_svc *svc, size_t *nr, mog_scandev_cb cb)
{
	struct dirent *ent;
	int rc = 0;

	if (svc->mgmt_fd < 0)
		return 0;

	CHECK(int, 0, pthread_mutex_lock(&svc->devstats_lock));
	svc_init_dev_hash(svc);
	rewinddir(svc->dir);
	while ((ent = readdir(svc->dir))) {
		unsigned long mog_devid;
		char *end;
		size_t len = strlen(ent->d_name);
		struct mog_dev *dev;

		if (len <= 3) continue;
		if (memcmp("dev", ent->d_name, 3) != 0) continue;

		mog_devid = strtoul(ent->d_name + 3, &end, 10);
		if (*end != 0) continue;
		if (mog_devid > 0xffffff) continue; /* MEDIUMINT in DB */

		dev = mog_dev_new(svc, (uint32_t)mog_devid);
		if (!dev) continue;

		if (cb) rc |= cb(dev, svc); /* mog_dev_mkusage */
		switch (hash_insert_if_absent(svc->by_st_dev, dev, NULL)) {
		case 0:
			free(dev);
			break;
		case 1:
			(*nr)++;
			break;
		default: mog_oom(); /* -1 */
		}
	}
	CHECK(int, 0, pthread_mutex_unlock(&svc->devstats_lock));

	return rc;
}

static bool write_devstats(void *entry, void *file)
{
	struct mog_dev *dev = entry;
	FILE *fp = file;
	char util[8];
	char *found = mog_mnt_fetch_util(dev->st_dev, util, sizeof(util));

	if (found)
		fprintf(fp, "%u\t%s\n", dev->devid, found);
	else
		fprintf(fp, "%u\t-\n", dev->devid);
	return true;
}

/* updates per-svc device stats from the global mount list */
static size_t devstats_stringify(struct mog_svc *svc, char **dst)
{
	FILE *fp;
	size_t bytes;

	assert(svc->by_st_dev && "need to scan devices first");

	fp = open_memstream(dst, &bytes);
	if (!fp) mog_oom();

	hash_do_for_each(svc->by_st_dev, write_devstats, fp);
	fprintf(fp, ".\n");

	CHECK(int, 0, my_memstream_close(fp, dst, &bytes));

	return bytes;
}

void mog_svc_devstats_subscribe(struct mog_mgmt *mgmt)
{
	struct mog_svc *svc = mgmt->svc;

	CHECK(int, 0, pthread_mutex_lock(&svc->devstats_lock));
	LIST_INSERT_HEAD(&svc->devstats_subscribers, mgmt, subscribed);
	CHECK(int, 0, pthread_mutex_unlock(&svc->devstats_lock));
}

/* called while iterating through all mog_svc objects */
bool mog_svc_devstats_broadcast(void *ent, void *ignored)
{
	struct mog_svc *svc = ent;
	struct mog_mgmt *mgmt, *tmp;
	struct iovec iov;
	char *buf = NULL;
	size_t len;
	struct mog_fd *mfd;

	CHECK(int, 0, pthread_mutex_lock(&svc->devstats_lock));

	len = devstats_stringify(svc, &buf);
	LIST_FOREACH_SAFE(mgmt, &svc->devstats_subscribers, subscribed, tmp) {

		assert(mgmt->wbuf == NULL && "wbuf not null");
		iov.iov_base = buf;
		iov.iov_len = len;
		mog_mgmt_writev(mgmt, &iov, 1);

		if (mgmt->wbuf == NULL) continue; /* success */

		LIST_REMOVE(mgmt, subscribed);
		mfd = mog_fd_of(mgmt);
		if (mgmt->wbuf == MOG_WR_ERROR) {
			assert(mgmt->rbuf == NULL && "would leak rbuf");
			mog_fd_put(mfd);
		} else { /* blocked on write */
			mog_idleq_push(mgmt->svc->queue, mfd, MOG_QEV_WR);
		}
	}
	free(buf);

	CHECK(int, 0, pthread_mutex_unlock(&svc->devstats_lock));

	return true;
}

static bool devstats_shutdown_i(void *svcptr, void *ignored)
{
	struct mog_svc *svc = svcptr;
	struct mog_mgmt *mgmt, *tmp;
	struct mog_fd *mfd;

	CHECK(int, 0, pthread_mutex_lock(&svc->devstats_lock));
	LIST_FOREACH_SAFE(mgmt, &svc->devstats_subscribers, subscribed, tmp) {
		assert(mgmt->wbuf == NULL && "wbuf not null");
		assert(mgmt->rbuf == NULL && "would leak rbuf");
		LIST_REMOVE(mgmt, subscribed);
		mfd = mog_fd_of(mgmt);
		mog_fd_put(mfd);
	}
	CHECK(int, 0, pthread_mutex_unlock(&svc->devstats_lock));

	return true;
}

void mog_svc_dev_shutdown(void)
{
	mog_svc_each(devstats_shutdown_i, NULL);
}

static bool svc_mkusage_each(void *svc, void *nr)
{
	svc_scandev((struct mog_svc *)svc, nr, mog_dev_mkusage);

	return true;
}

size_t mog_mkusage_all(void)
{
	size_t nr = 0;

	mog_svc_each(svc_mkusage_each, &nr);

	return nr;
}
